"""
.. include:: ../README.md
.. include:: ../docs/user_guide/1_concepts.md
.. include:: ../docs/user_guide/2_credentials.md
.. include:: ../docs/user_guide/3_postprocess.md
.. include:: ../docs/user_guide/4_pipelines.md
.. include:: ../docs/developer_guide/1_main.md
"""
